from getpass import getpass
import os
import os.path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import click

import lxml.etree as et

from ncclient import manager

from tqdm import tqdm


cfg = dict()


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-c", "--config", type=click.File("rb"), help="Config file")
def main(config):
    if config is not None:
        cfg.update(tomllib.load(config))


@main.command()
@click.option("-D", "--device", help="Device defined in config")
@click.option("-a", "--host", help="Device IP address or Hostname")
@click.option(
    "--port", type=int, default=830, show_default=True, help="Netconf agent port"
)
@click.option("-u", "--username", help="Device Username (netconf agent username)")
@click.option(
    "-d",
    "--destination",
    default="models",
    type=click.Path(file_okay=False),
    help="Destination directory",
)
@click.pass_context
def get_models(ctx: click.Context, device, host, port, username, destination):
    """Get YANG models from device"""

    dev = cfg.get("devices", {}).get(device, {})
    _host = host or dev["host"]
    if ctx.get_parameter_source("port").name() == "COMMANDLINE":
        _port = port
    else:
        _port = dev.get("port", port)
    _username = username or dev["username"]

    try:
        password = getpass(prompt="Password (C-d for none/agent only): ")
    except EOFError:
        password = None

    with manager.connect(
        host=_host,
        port=_port,
        username=_username,
        password=password,
        timeout=90,
        device_params=dev.get("params"),
    ) as m:
        caps = list(map(lambda x: x.strip(), m.server_capabilities))
        if not any(map(lambda x: x.startswith(manager.NETCONF_MONITORING_NS), caps)):
            raise click.ClickException("device does not support rfc6022")

        _filter = et.fromstring(
            f"""
            <netconf-state xmlns="{manager.NETCONF_MONITORING_NS}">
              <schemas/>
            </netconf-state>"""
        )
        res = m.get(("subtree", _filter))

        prefix = f"{{{manager.NETCONF_MONITORING_NS}}}"

        os.makedirs(destination, exist_ok=True)

        pbar = tqdm(
            res.data.findall(
                "./x:netconf-state/x:schemas/x:schema",
                {"x": manager.NETCONF_MONITORING_NS},
            )
        )
        for schema in pbar:
            identifier = schema.findtext(prefix + "identifier")
            assert identifier is not None
            version = schema.findtext(prefix + "version")
            fmt = schema.findtext(prefix + "format")
            assert any(
                [el.text == "NETCONF" for el in schema.findall(prefix + "location")]
            )
            if version is not None and version != "":
                v = f"@{version}"
            else:
                v = ""
            filename = f"{identifier}{v}.{fmt}"
            # pbar.set_description(filename)

            # Don’t pass format because IOS XE doesn’t eat its own definition.
            r = m.get_schema(identifier, version)

            with open(os.path.join(destination, filename), "w") as fp:
                fp.write(r.data)


@main.command()
@click.option("-D", "--device", help="Device defined in config")
@click.option("-o", "--output", type=click.File("w"), default="-")
@click.option(
    "--source",
    type=click.Choice(["running", "startup", "candidate"]),
    default="running",
    help="Config store",
)
@click.argument("xpath")
def get_config(device, output, source, xpath):
    """Get config from device"""

    dev = cfg["devices"][device]

    try:
        password = getpass(prompt="Password (C-d for none/agent only): ")
    except EOFError:
        password = None

    with manager.connect(
        host=dev.get("hostname"),
        port=dev.get("port", 830),  # TODO
        username=dev.get("user"),
        password=password,
        timeout=90,
        device_params=dev.get("params"),
    ) as m:
        res = m.get_config(source, filter=("xpath", (dev.get("namespaces", {}), xpath)))
        output.write(res.data_xml)
